from ...models import DocumentTypeMetadataType

from .metadata_type_mixins import MetadataTypeTestMixin


class DocumentTypeMetadataTypeTestMixin(MetadataTypeTestMixin):
    _test_object_model = DocumentTypeMetadataType
    _test_object_name = '_test_document_type_metadata_type'

    def _create_test_document_type_metadata_type(self):
        self._test_document_type_metadata_type = self._test_document_type.metadata.create(
            metadata_type=self._test_metadata_type, required=False
        )


class DocumentTypeMetadataTypeAPIViewTestMixin(
    DocumentTypeMetadataTypeTestMixin
):
    def _request_document_type_metadata_type_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:documenttypemetadatatype-list',
            kwargs={'document_type_id': self._test_document_type.pk}, data={
                'metadata_type_id': self._test_metadata_type.pk,
                'required': False
            }
        )

        self._test_object_set()

        return response

    def _request_document_type_metadata_type_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documenttypemetadatatype-detail',
            kwargs={
                'document_type_id': self._test_document_type.pk,
                'metadata_type_id': self._test_document_type_metadata_type.pk
            }
        )

    def _request_document_type_metadata_type_list_api_view(self):
        return self.get(
            viewname='rest_api:documenttypemetadatatype-list', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )

    def _request_document_type_metadata_type_edit_api_view_via_patch(self):
        return self.patch(
            viewname='rest_api:documenttypemetadatatype-detail',
            kwargs={
                'document_type_id': self._test_document_type.pk,
                'metadata_type_id': self._test_document_type_metadata_type.pk
            }, data={
                'required': True
            }
        )

    def _request_document_type_metadata_type_edit_api_view_via_put(self):
        return self.put(
            viewname='rest_api:documenttypemetadatatype-detail',
            kwargs={
                'document_type_id': self._test_document_type.pk,
                'metadata_type_id': self._test_document_type_metadata_type.pk
            }, data={
                'required': True
            }
        )



from mayan.apps.sources.tests.literals import TEST_CASE_ACTION_NAME_SOURCE_CREATE
from mayan.apps.sources.tests.mixins.base_mixins import SourceTestMixin

from ..source_backends.literals import DEFAULT_PERIOD_INTERVAL

from .literals import TEST_SOURCE_BACKEND_PATH_PERIODIC


class PeriodicSourceBackendTestMixin(SourceTestMixin):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_PERIODIC

    def _get_test_source_backend_data(self, action_name, interface_name):
        result = super()._get_test_source_backend_data(
            action_name=action_name, interface_name=interface_name
        )

        if action_name == TEST_CASE_ACTION_NAME_SOURCE_CREATE:
            result.update(
                {
                    'document_type_id': self._test_document_type.pk,
                    'interval': DEFAULT_PERIOD_INTERVAL
                }
            )

        return result



from .events import event_web_link_edited


def method_document_type_web_links_add(self, queryset, user):
    for model_instance in queryset:
        self.web_links.add(model_instance)
        event_web_link_edited.commit(
            action_object=self, actor=user, target=model_instance
        )


def method_document_type_web_links_remove(self, queryset, user):
    for model_instance in queryset:
        self.web_links.remove(model_instance)
        event_web_link_edited.commit(
            action_object=self, actor=user, target=model_instance
        )



from django.http import Http404
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import SingleObjectListView

from .classes import CeleryQueue, Worker
from .icons import (
    icon_queue_task_type_list, icon_worker_list, icon_worker_queue_list
)
from .permissions import permission_task_view


class QueueTaskTypeListView(SingleObjectListView):
    view_icon = icon_queue_task_type_list
    view_permission = permission_task_view

    def get_extra_context(self):
        queue = self.get_queue()

        return {
            'hide_object': True,
            'object': queue,
            'worker': queue.worker,
            'navigation_object_list': ('worker', 'object'),
            'title': _(message='Task types for queue: %s') % queue
        }

    def get_queue(self):
        try:
            return CeleryQueue.get(
                queue_name=self.kwargs['queue_name']
            )
        except KeyError:
            raise Http404(
                _(message='Queue: %s, not found') % self.kwargs['queue_name']
            )

    def get_source_queryset(self):
        return self.get_queue().task_types


class WorkerListView(SingleObjectListView):
    extra_context = {
        'hide_object': True,
        'title': _(message='Background task workers')
    }
    view_icon = icon_worker_list
    view_permission = permission_task_view

    def get_source_queryset(self):
        return Worker.all()


class WorkerQueueListView(SingleObjectListView):
    view_icon = icon_worker_queue_list
    view_permission = permission_task_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'object': self.get_worker(),
            'title': _(message='Queues for worker: %s') % self.get_worker()
        }

    def get_worker(self):
        try:
            return Worker.get(
                name=self.kwargs['worker_name']
            )
        except KeyError:
            raise Http404(
                _(message='Worker: %s, not found') % self.kwargs['worker_name']
            )

    def get_source_queryset(self):
        return self.get_worker().queues



#!/usr/bin/env python

import doctest
import sys

from packaging.version import Version as PackageVersion


class Version:
    """
    Subset of PEP 440
    N(.N)*[{a|b|rc}N][.postN][.devN][+<local version label>]

    Release segment: N(.N)*
    Pre-release segment: {a|b|rc}N
    Post-release segment: .postN
    Development release segment: .devN
    Local version: +local version label

    # Return value

    >>> Version('1')
    Version: 1
    >>> Version('1.0')
    Version: 1.0
    >>> Version('1.3.2')
    Version: 1.3.2
    >>> Version('1.3.2dev0')
    Version: 1.3.2.dev0
    >>> Version('1.3.dev0')
    Version: 1.3.dev0
    >>> Version('1.3.2+local.1')
    Version: 1.3.2+local.1

    >>> Version('1.2.3').major
    1
    >>> Version('1.2.3').minor
    2
    >>> Version('1.2.3').micro
    3
    >>> Version('1.2.3').dev

    >>> Version('1.2.3dev0').dev
    '0'
    >>> Version('1.2.3.dev').dev
    '0'
    >>> Version('1.2.3.dev0').dev
    '0'
    >>> Version('1.2.3').pre

    >>> Version('1.2.3.a0').pre
    'a0'
    >>> Version('1.2.3a').pre
    'a0'
    >>> Version('1.2.3a0').pre
    'a0'
    >>> Version('1.2.3.b0').pre
    'b0'
    >>> Version('1.2.3b').pre
    'b0'
    >>> Version('1.2.3b0').pre
    'b0'
    >>> Version('1.2.3.rc0').pre
    'rc0'
    >>> Version('1.2.3rc').pre
    'rc0'
    >>> Version('1.2.3rc0').pre
    'rc0'
    >>> Version('1.2.3').post

    >>> Version('1.2.3.post').post
    '0'
    >>> Version('1.2.3.post0').post
    '0'
    >>> Version('1.2.3.post1').post
    '1'

    # Version part increment

    >>> Version('1.2.3').increment_major()
    Version: 2
    >>> Version('1.2.3').increment_minor()
    Version: 1.3
    >>> Version('1.2.3').increment_micro()
    Version: 1.2.4
    >>> Version('1.2.3dev0').increment_micro()
    Version: 1.2.4
    >>> Version('1.2.3').increment_dev()
    Version: 1.2.3.dev0
    >>> Version('1.2.3dev0').increment_dev()
    Version: 1.2.3.dev1
    >>> Version('1.2.3').increment_pre()
    Version: 1.2.3a0
    >>> Version('1.2.3a0').increment_pre()
    Version: 1.2.3a1
    >>> Version('1.2.3b').increment_pre()
    Version: 1.2.3b1
    >>> Version('1.2.3b0').increment_pre()
    Version: 1.2.3b1
    >>> Version('1.2.3rc').increment_pre()
    Version: 1.2.3rc1
    >>> Version('1.2.3rc0').increment_pre()
    Version: 1.2.3rc1
    >>> Version('1.2.3').increment_post()
    Version: 1.2.3.post0
    >>> Version('1.2.3.post0').increment_post()
    Version: 1.2.3.post1

    # Unofficial increments

    >>> Version('1.2.3a1.post0').increment_post()
    Version: 1.2.3a1.post1
    >>> Version('1.2.3b1.post0').increment_post()
    Version: 1.2.3b1.post1
    >>> Version('1.2.3rc1.post0').increment_post()
    Version: 1.2.3rc1.post1
    >>> Version('1.2.3rc1.post0+local.1').increment_post()
    Version: 1.2.3rc1.post1+local.1

    # Increment part, local version

    >>> Version('1.2.3+local.1').increment_major()
    Version: 2+local.1
    >>> Version('1.2.3+local.1').increment_minor()
    Version: 1.3+local.1
    >>> Version('1.2.3+local.1').increment_micro()
    Version: 1.2.4+local.1
    >>> Version('1.2.3dev0+local.1').increment_micro()
    Version: 1.2.4+local.1

    # As a part

    >>> Version('1.2.3').as_major()
    '1'
    >>> Version('1.2.3').as_minor()
    '1.2'
    >>> Version('1.2.3').as_micro()
    '1.2.3'
    >>> Version('1.2').as_micro()
    '1.2.0'

    # As a part, local version

    >>> Version('1.2.3+local.1').as_major()
    '1+local.1'
    >>> Version('1.2.3+local.1').as_minor()
    '1.2+local.1'
    >>> Version('1.2.3+local.1').as_micro()
    '1.2.3+local.1'
    >>> Version('1.2+local.1').as_micro()
    '1.2.0+local.1'
    >>> Version('1+local.1').as_minor()
    '1.0+local.1'

    # As a part, local version, as upstream

    >>> Version('1+local.1').as_upstream()
    '1'
    >>> Version('1.2+local.1').as_upstream()
    '1.2'
    >>> Version('1.2.3+local.1').as_upstream()
    '1.2.3'
    >>> Version('1.2.3dev0+local.1').as_upstream()
    '1.2.3.dev0'

    # Comparison, greater than

    >>> Version('1') > Version('2')
    False
    >>> Version('2') > Version('1')
    True
    >>> Version('1.1') > Version('1.2')
    False
    >>> Version('1.2') > Version('1.1')
    True
    >>> Version('1.1.2') > Version('1.1.1')
    True
    >>> Version('1.1.1') > Version('1.1.2')
    False
    >>> Version('1.1.1dev0') > Version('1.1.2dev0')
    False
    >>> Version('1.1.1dev1') > Version('1.1.2dev0')
    False
    >>> Version('1.1.2dev0') > Version('1.1.1dev1')
    True
    >>> Version('1.1.1dev1') > Version('1.1.1dev0')
    True
    >>> Version('1.1') > Version('2')
    False
    >>> Version('1.1') > Version('1')
    True
    >>> Version('1') > Version('1.1')
    False

    # Comparison, greater than, different local, same upstream

    >>> Version('1') > Version('1+local.1')
    False
    >>> Version('1+local.1') > Version('1')
    True
    >>> Version('1.2') > Version('1.2+local.1')
    False
    >>> Version('1.2+local.1') > Version('1.2')
    True
    >>> Version('1.1.1') > Version('1.1.1+local.1')
    False
    >>> Version('1.1.1+local.1') > Version('1.1.1')
    True

    # Comparison, greater than, same local, different upstream

    >>> Version('1+local.1') > Version('2+local.1')
    False
    >>> Version('2+local.1') > Version('1+local.1')
    True
    >>> Version('1.1+local.1') > Version('1.2+local.1')
    False
    >>> Version('1.2+local.1') > Version('1.1+local.1')
    True
    >>> Version('1.1.1+local.1') > Version('1.1.2+local.1')
    False
    >>> Version('1.1.2+local.1') > Version('1.1.1+local.1')
    True

    # Comparison, greater than, different local, different upstream

    >>> Version('1+local.1') > Version('2')
    False
    >>> Version('2') > Version('1+local.1')
    True
    >>> Version('1+local.1') > Version('1.2')
    False
    >>> Version('1.1') > Version('1+local.1')
    True
    >>> Version('1+local.1') > Version('1.1')
    False
    >>> Version('1.2') > Version('1+local.1')
    True
    >>> Version('1+local.1') > Version('1.2')
    False
    >>> Version('1.1+local.1') > Version('1.2')
    False
    >>> Version('1.2') > Version('1.1+local.1')
    True
    >>> Version('1.1.1+local.1') > Version('1.1.2')
    False
    >>> Version('1.1.2') > Version('1.1.1+local.1')
    True
    >>> Version('1.1.2+local.1') > Version('1.1.1')
    True
    >>> Version('1.1.1') > Version('1.1.2+local.1')
    False

    # Comparison, less than

    >>> Version('1') < Version('2')
    True
    >>> Version('2') < Version('1')
    False
    >>> Version('1.1') < Version('1.2')
    True
    >>> Version('1.2') < Version('1.1')
    False
    >>> Version('1.1.2') < Version('1.1.1')
    False
    >>> Version('1.1.1') < Version('1.1.2')
    True
    >>> Version('1.1') < Version('2')
    True
    >>> Version('1.1') < Version('1')
    False
    >>> Version('1') < Version('1.1')
    True

    # Comparison, greater than, different local, same upstream

    >>> Version('1') < Version('1+local.1')
    True
    >>> Version('1+local.1') < Version('1')
    False
    >>> Version('1.2') < Version('1.2+local.1')
    True
    >>> Version('1.2+local.1') < Version('1.2')
    False
    >>> Version('1.1.1') < Version('1.1.1+local.1')
    True
    >>> Version('1.1.1+local.1') < Version('1.1.1')
    False

    # Comparison, greater than, same local, different upstream

    >>> Version('1+local.1') < Version('2+local.1')
    True
    >>> Version('2+local.1') < Version('1+local.1')
    False
    >>> Version('1.1+local.1') < Version('1.2+local.1')
    True
    >>> Version('1.2+local.1') < Version('1.1+local.1')
    False
    >>> Version('1.1.1+local.1') < Version('1.1.2+local.1')
    True
    >>> Version('1.1.2+local.1') < Version('1.1.1+local.1')
    False

    # Comparison, greater than, different local, different upstream

    >>> Version('1+local.1') < Version('2')
    True
    >>> Version('2') < Version('1+local.1')
    False
    >>> Version('1+local.1') < Version('1.2')
    True
    >>> Version('1.1') < Version('1+local.1')
    False
    >>> Version('1+local.1') < Version('1.1')
    True
    >>> Version('1.2') < Version('1+local.1')
    False
    >>> Version('1+local.1') < Version('1.2')
    True
    >>> Version('1.1+local.1') < Version('1.2')
    True
    >>> Version('1.2') < Version('1.1+local.1')
    False
    >>> Version('1.1.1+local.1') < Version('1.1.2')
    True
    >>> Version('1.1.2') < Version('1.1.1+local.1')
    False
    >>> Version('1.1.2+local.1') < Version('1.1.1')
    False
    >>> Version('1.1.1') < Version('1.1.2+local.1')
    True
    """
    def __eq__(self, other):
        return self._version == other._version

    def __gt__(self, other):
        return self._version > other._version

    def __lt__(self, other):
        return self._version < other._version

    def __init__(self, version_string):
        self._version_string = version_string
        self._version = PackageVersion(version=self._version_string)

    def __repr__(self):
        return 'Version: {}'.format(self._version)

    def _as_base(self, parts):
        result = '.'.join(
            map(
                str, parts
            )
        )

        return result

    def _as_add_suffixes(self, version_string):
        result = '{}'.format(version_string)

        if self._version.local:
            result = '{}+{}'.format(result, self._version.local)

        return result

    def as_major(self):
        base = self._as_base(
            parts=(self._version.major,)
        )

        return self._as_add_suffixes(version_string=base)

    def as_micro(self):
        base = self._as_base(
            parts=(
                self._version.major, self._version.minor, self._version.micro
            )
        )

        return self._as_add_suffixes(version_string=base)

    def as_minor(self):
        base = self._as_base(
            parts=(self._version.major, self._version.minor)
        )

        return self._as_add_suffixes(version_string=base)

    def as_upstream(self):
        return self._version.public

    @property
    def dev(self):
        if self._version.dev is not None:
            return str(self._version.dev)

    def get_version_string(self):
        return str(self._version)

    def increment_dev(self):
        version_string = '.'.join(
            map(str, self._version.release)
        )

        if self._version.dev is None:
            dev = -1
        else:
            dev = self._version.dev or 0

        version_string = '{}.dev{}'.format(version_string, dev + 1)

        if self._version.local:
            version_string = '{}+{}'.format(
                version_string, self._version.local
            )

        self._version = PackageVersion(
            version=version_string
        )
        return self

    def increment_major(self):
        version_string = '{}'.format(
            self._version.major + 1
        )

        if self._version.local:
            version_string = '{}+{}'.format(
                version_string, self._version.local
            )

        self._version = PackageVersion(
            version=version_string
        )
        return self

    def increment_micro(self):
        version_string = '{}.{}.{}'.format(
            self._version.major, self._version.minor, self._version.micro + 1
        )

        if self._version.local:
            version_string = '{}+{}'.format(
                version_string, self._version.local
            )

        self._version = PackageVersion(
            version=version_string
        )
        return self

    def increment_minor(self):
        version_string = '{}.{}'.format(
            self._version.major, self._version.minor + 1
        )

        if self._version.local:
            version_string = '{}+{}'.format(
                version_string, self._version.local
            )

        self._version = PackageVersion(
            version=version_string
        )
        return self

    def increment_pre(self):
        version_string = '.'.join(
            map(str, self._version.release)
        )

        if self._version.pre is None:
            pre = ('a', -1)
        else:
            pre = self._version.pre

        version_string = '{}{}{}'.format(version_string, pre[0], pre[1] + 1)

        if self._version.local:
            version_string = '{}+{}'.format(
                version_string, self._version.local
            )

        self._version = PackageVersion(
            version=version_string
        )
        return self

    def increment_post(self):
        version_string = '.'.join(
            map(str, self._version.release)
        )

        if self._version.pre is not None:
            version_string = '{}{}{}'.format(
                version_string, self._version.pre[0], self._version.pre[1]
            )

        if self._version.post is None:
            post = -1
        else:
            post = self._version.post or 0

        version_string = '{}.post{}'.format(version_string, post + 1)

        if self._version.local:
            version_string = '{}+{}'.format(
                version_string, self._version.local
            )

        self._version = PackageVersion(
            version=version_string
        )
        return self

    @property
    def major(self):
        return self._version.major

    @property
    def micro(self):
        return self._version.micro

    @property
    def minor(self):
        return self._version.minor

    @property
    def post(self):
        if self._version.post is not None:
            return str(self._version.post)

    @property
    def pre(self):
        if self._version.pre:
            return ''.join(
                map(
                    str, self._version.pre
                )
            )


if __name__ == '__main__':
    if len(sys.argv) == 2 and sys.argv[1] == '-test':
        doctest.testmod()
        exit(0)

    if len(sys.argv) == 1:
        print(
            'usage: [version] <part to retrieve [major, minor, micro]> <-test>'
        )
        exit(0)

    if len(sys.argv) < 3:
        print('Insufficient arguments')
        exit(1)

    version_string = sys.argv[1]
    if version_string == '-':
        version_string = sys.stdin.read().replace('\n', '')

    version = Version(version_string=version_string)
    part = sys.argv[2].lower()

    if part == 'major':
        output = version.as_major()
    elif part == 'minor':
        output = version.as_minor() or ''
    elif part == 'micro':
        output = version.as_micro() or ''
    elif part == 'upstream':
        output = version.as_upstream() or ''
    else:
        print('Unknown part')
        exit(1)

    print(output)



from rest_framework import status
from rest_framework.response import Response

from mayan.apps.documents.serializers.document_serializers import DocumentSerializer
from mayan.apps.documents.tasks import task_document_file_create

from ..interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)

from .arguments import (
    argument_immediate_mode_optional, argument_immediate_mode_required
)
from .document_mixins import SourceBackendActionMixinDocumentOptionalTaskOnly
from .literals import DEFAULT_IMMEDIATE_MODE


class SourceBackendActionMixinImmediateMode(SourceBackendActionMixinDocumentOptionalTaskOnly):
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                immediate_mode = argument_immediate_mode_optional

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['immediate_mode'] = self.context['immediate_mode']

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                immediate_mode = argument_immediate_mode_optional

            def process_action_data(self):
                super().process_action_data()

                immediate_mode = self.action_kwargs.get(
                    'immediate_mode', DEFAULT_IMMEDIATE_MODE
                )

                if immediate_mode:
                    request = self.context['view'].request

                    serializer = DocumentSerializer(
                        context={'request': request},
                        instance=self.action_data
                    )

                    self.interface_result = Response(
                        data=serializer.data, status=status.HTTP_201_CREATED
                    )

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['immediate_mode'] = self.context['immediate_mode']

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                immediate_mode = argument_immediate_mode_required

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['immediate_mode'] = self.context['immediate_mode']

        class View(SourceBackendActionInterfaceRequestViewForm):
            def process_interface_context(self):
                super().process_interface_context()

                # Provide a static default for HTTP views. This way
                # a default does not need to be added in the method argument.
                # Avoids `(immediate_mode=DEFAULT_IMMEDIATE_MODE)`.
                self.action_kwargs['immediate_mode'] = DEFAULT_IMMEDIATE_MODE

    def _background_task(self, immediate_mode, **kwargs):
        result = super()._background_task(**kwargs)

        if immediate_mode:
            server_upload_entry = result['server_upload_entry_list'][0]

            document_file_task_kwargs = self.get_document_file_task_kwargs(
                server_upload_entry=server_upload_entry, **kwargs
            )
            document_file_task_kwargs['is_document_upload_sequence'] = True

            document_file_task_kwargs['shared_uploaded_file_id'] = server_upload_entry[
                'shared_uploaded_file_id'
            ]

            task_document_file_create.apply_async(
                kwargs=document_file_task_kwargs
            )
        else:
            return result

    def _execute(self, immediate_mode, **kwargs):
        if immediate_mode:
            self.document = kwargs['document_type'].documents_create(
                description=kwargs['description'], label=kwargs['label'],
                language=kwargs['language'], user=kwargs['user']
            )

            super()._execute(immediate_mode=immediate_mode, **kwargs)

            return self.document
        else:
            super()._execute(immediate_mode=immediate_mode, **kwargs)

    def get_task_kwargs(self, immediate_mode, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['immediate_mode'] = immediate_mode

        # Check if a document was create as part of the immediate mode.
        document = getattr(self, 'document', None)

        if document:
            result['action_interface_kwargs']['document_id'] = document.pk

        return result



from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import SimpleView, SingleObjectListView

from .classes import Dashboard
from .icons import icon_dashboard_detail, icon_dashboard_list


class DashboardListView(SingleObjectListView):
    extra_context = {
        'hide_object': True,
        'title': _(message='Dashboards')
    }
    view_icon = icon_dashboard_list

    def get_source_queryset(self):
        return Dashboard.get_all()


class DashboardDetailView(SimpleView):
    template_name = 'appearance/content_container.html'
    view_icon = icon_dashboard_detail

    def get_extra_context(self):
        dashboard = Dashboard.get(
            name=self.kwargs['dashboard_name']
        )

        return {
            'content': dashboard.render(request=self.request),
            'object': dashboard,
            'title': _(message='Dashboard detail')
        }



from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH
from mayan.apps.sources.tests.mixins.base_mixins import SourceTestMixin

from .literals import TEST_SOURCE_BACKEND_PATH_COMPRESSED


class CompressedSourceTestMixin(SourceTestMixin):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_COMPRESSED
    _test_source_file_path = TEST_FILE_SMALL_PATH



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import DEFAULT_AUTHENTICATION_OIDC_USER_PROFILE_URL

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Authentication OIDC'), name='authentication_oidc'
)

setting_oidc_user_profile_url = setting_namespace.do_setting_add(
    default=DEFAULT_AUTHENTICATION_OIDC_USER_PROFILE_URL,
    global_name='AUTHENTICATION_OIDC_USER_PROFILE_URL',
    help_text=_(message='URL for the OIDC user profile page.')
)



from mayan.apps.converter.api_view_mixins import APIImageViewMixin
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_view
from mayan.apps.documents.serializers.document_type_serializers import DocumentTypeSerializer
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from ..models.workflow_models import Workflow
from ..permissions import (
    permission_workflow_template_create,
    permission_workflow_template_delete, permission_workflow_template_edit,
    permission_workflow_template_view
)
from ..serializers.workflow_template_serializers import (
    WorkflowTemplateDocumentTypeAddSerializer,
    WorkflowTemplateDocumentTypeRemoveSerializer, WorkflowTemplateSerializer
)


class APIWorkflowTemplateDocumentTypeListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the document types attached to a workflow template.
    """
    external_object_class = Workflow
    external_object_pk_url_kwarg = 'workflow_template_id'
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view
    }
    mayan_object_permission_map = {'GET': permission_document_type_view}
    serializer_class = DocumentTypeSerializer

    def get_source_queryset(self):
        """
        This view returns a list of document types that belong to a workflow template.
        """
        return self.get_external_object().document_types.all()


class APIWorkflowTemplateDocumentTypeAddView(generics.ObjectActionAPIView):
    """
    post: Add a document type to a workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permission_map = {
        'POST': permission_workflow_template_edit
    }
    serializer_class = WorkflowTemplateDocumentTypeAddSerializer
    source_queryset = Workflow.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type_id']
        obj.document_types_add(
            queryset=DocumentType.objects.filter(pk=document_type.id),
            user=self.request.user
        )


class APIWorkflowTemplateDocumentTypeRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document type from a workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permission_map = {
        'POST': permission_workflow_template_edit
    }
    serializer_class = WorkflowTemplateDocumentTypeRemoveSerializer
    source_queryset = Workflow.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type_id']
        obj.document_types_remove(
            queryset=DocumentType.objects.filter(pk=document_type.id),
            user=self.request.user
        )


class APIWorkflowTemplateImageView(
    APIImageViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permission_map = {'GET': permission_workflow_template_view}
    source_queryset = Workflow.objects.all()


class APIWorkflowTemplateListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the workflow templates.
    post: Create a new workflow template.
    """
    mayan_object_permission_map = {'GET': permission_workflow_template_view}
    mayan_view_permission_map = {
        'POST': permission_workflow_template_create
    }
    ordering_fields = ('id', 'internal_name', 'label')
    serializer_class = WorkflowTemplateSerializer
    source_queryset = Workflow.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIWorkflowTemplateDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected workflow template.
    get: Return the details of the selected workflow template.
    patch: Edit the selected workflow template.
    put: Edit the selected workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permission_map = {
        'DELETE': permission_workflow_template_delete,
        'GET': permission_workflow_template_view,
        'PATCH': permission_workflow_template_edit,
        'PUT': permission_workflow_template_edit
    }
    serializer_class = WorkflowTemplateSerializer
    source_queryset = Workflow.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_tag_attached, event_tag_removed
from ..permissions import (
    permission_tag_attach, permission_tag_remove, permission_tag_view
)

from .mixins import TagAPIViewTestMixin, TagTestMixin


class DocumentTagAPIViewTestCase(
    DocumentTestMixin, TagAPIViewTestMixin, TagTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_tag()
        self._create_test_document_stub()

    def test_document_attach_tag_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_tag_attach_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_attach_tag_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_tag_attach_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_attach_tag_api_view_with_tag_access(self):
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_tag_attach_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_attach_tag_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_tag_attach_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_attached.id)

    def test_trashed_document_attach_tag_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_tag_attach_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_list_api_view_no_permission(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self._clear_events()

        response = self._request_test_document_tag_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_list_api_view_with_document_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_document, permission=permission_tag_view
        )

        self._clear_events()

        response = self._request_test_document_tag_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_list_api_view_with_tag_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._clear_events()

        response = self._request_test_document_tag_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_list_api_view_with_full_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_document, permission=permission_tag_view
        )
        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._clear_events()

        response = self._request_test_document_tag_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['label'], self._test_tag.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_tag_list_api_view_with_full_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_document, permission=permission_tag_view
        )
        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_tag_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_remove_api_view_no_permission(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self._clear_events()

        response = self._request_test_document_tag_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_remove_api_view_with_document_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_tag_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_remove_api_view_with_tag_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_tag_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_remove_api_view_with_full_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_tag_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertFalse(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_removed.id)

    def test_trashed_document_tag_remove_api_view_with_full_access(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_tag_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from unittest import mock

from mayan.apps.lock_manager.exceptions import LockError
from mayan.apps.testing.tests.base import BaseTestCase

from ..events import (
    event_cache_created, event_cache_edited, event_cache_partition_purged,
    event_cache_purged
)
from ..exceptions import FileCachingException
from ..models import CachePartitionFile

from .literals import TEST_CACHE_PARTITION_FILE_FILENAME
from .mixins import CacheTestMixin


class CacheModelTestCase(CacheTestMixin, BaseTestCase):
    class FakeException(Exception):
        """
        Exception to force the cache file creation to fail but not the
        test itself.
        """

    def test_cache_create(self):
        self._clear_events()

        self._create_test_cache()

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_cache)
        self.assertEqual(events[0].target, self._test_cache)
        self.assertEqual(events[0].verb, event_cache_created.id)

    def test_cache_get_absolute_url_method(self):
        self._create_test_cache()

        self._clear_events()

        self.assertTrue(
            self._test_cache.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_purge(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        cache_total_size = self._test_cache.get_total_size()

        self._clear_events()
        self._test_cache.purge(user=self._test_case_user)

        self.assertNotEqual(
            cache_total_size, self._test_cache.get_total_size()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_cache)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_cache_partition)
        self.assertEqual(events[0].verb, event_cache_partition_purged.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_cache)
        self.assertEqual(events[1].verb, event_cache_purged.id)

    @mock.patch('django.core.files.File.close')
    def test_storage_file_close(self, mock_storage_file_close_method):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        self._clear_events()

        self.assertTrue(mock_storage_file_close_method.called)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cleanup_on_storage_file_creation_error(self):
        self._silence_logger(name='mayan.apps.file_caching.model_mixins')

        self._create_test_cache()
        self._create_test_cache_partition()

        cache_parition_file_count = self._test_cache_partition.files.count()

        self._clear_events()

        with self.assertRaises(expected_exception=CacheModelTestCase.FakeException):
            with self._test_cache_partition.create_file(filename=TEST_CACHE_PARTITION_FILE_FILENAME):
                raise CacheModelTestCase.FakeException

        self.assertEqual(
            self._test_cache_partition.files.count(),
            cache_parition_file_count
        )
        self.assertFalse(
            self._test_cache_partition.cache.storage.exists(
                name=self._test_cache_partition.get_full_filename(
                    filename=TEST_CACHE_PARTITION_FILE_FILENAME
                )
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    @mock.patch('mayan.apps.file_caching.models.CachePartitionFile.save')
    def test_cleanup_on_model_file_creation_error(self, mock_save):
        self._silence_logger(name='mayan.apps.file_caching.model_mixins')

        mock_save.side_effect = CacheModelTestCase.FakeException

        self._create_test_cache()
        self._create_test_cache_partition()

        cache_parition_file_count = self._test_cache_partition.files.count()

        self._clear_events()

        with self.assertRaises(expected_exception=CacheModelTestCase.FakeException):
            with self._test_cache_partition.create_file(filename=TEST_CACHE_PARTITION_FILE_FILENAME) as file_object:
                file_object.write(b'')

        self.assertEqual(
            self._test_cache_partition.files.count(),
            cache_parition_file_count
        )
        self.assertFalse(
            self._test_cache_partition.cache.storage.exists(
                name=self._test_cache_partition.get_full_filename(
                    filename=TEST_CACHE_PARTITION_FILE_FILENAME
                )
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_partition_file_hits(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        cache_partition_file_hits = self._test_cache_partition_file.hits

        self._clear_events()

        with self._test_cache_partition_file.open():
            """Do nothing"""

        self._test_cache_partition_file.refresh_from_db()

        self.assertEqual(
            self._test_cache_partition_file.hits,
            cache_partition_file_hits + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_partition_file_lru_eviction(self):
        self._create_test_cache(
            extra_data={
                'maximum_size': 2
            }
        )

        self._create_test_cache_partition()
        self._create_test_cache_partition_file(file_size=1)
        self._create_test_cache_partition_file(file_size=1)

        self._clear_events()

        with self._test_cache_partition_files[0].open():
            """Do nothing"""

        self._create_test_cache_partition_file(file_size=1)

        # Older but more hits was kept.
        self.assertTrue(
            self._test_cache_partition_files[0] in CachePartitionFile.objects.all()
        )
        # Newer but less hits was purged.
        self.assertTrue(
            self._test_cache_partition_files[1] not in CachePartitionFile.objects.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_partition_file_size_protection(self):
        self._create_test_cache(
            extra_data={
                'maximum_size': 1
            }
        )

        self._create_test_cache_partition()

        self._clear_events()

        with self.assertRaises(expected_exception=FileCachingException):
            self._create_test_cache_partition_file(file_size=2)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    @mock.patch('mayan.apps.file_caching.models.Cache.prune')
    def test_prune_on_cache_size_reduction(self, mock_cache_prune_method):
        self._create_test_cache(
            extra_data={
                'maximum_size': 2
            }
        )

        self._clear_events()

        self._test_cache.maximum_size = 2
        self._test_cache.save()
        self.assertFalse(mock_cache_prune_method.called)

        self._test_cache.maximum_size = 1
        self._test_cache.save()
        self.assertTrue(mock_cache_prune_method.called)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_cache)
        self.assertEqual(events[0].target, self._test_cache)
        self.assertEqual(events[0].verb, event_cache_edited.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_cache)
        self.assertEqual(events[1].target, self._test_cache)
        self.assertEqual(events[1].verb, event_cache_edited.id)

    def test_cache_partition_file_context_manager_locking(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        self._clear_events()

        with self._test_cache_partition_file.open():
            with self.assertRaises(expected_exception=LockError):
                with self._test_cache_partition_file.open():
                    """Trigger LockError."""

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_incremental_file_index_cache_prune(self):
        self._create_test_cache(
            extra_data={
                'maximum_size': 2
            }
        )

        self._create_test_cache_partition()
        self._create_test_cache_partition_file(file_size=1)
        self._create_test_cache_partition_file(file_size=1)

        self._clear_events()

        with self._test_cache_partition_files[1].open():
            """Increase hits of file #1"""

        with self._test_cache_partition_files[0].open():
            """Lock and increase hits of file #0"""
            self._create_test_cache_partition_file(file_size=1)

        self.assertTrue(
            self._test_cache_partition_files[0] in CachePartitionFile.objects.all()
        )
        self.assertTrue(
            self._test_cache_partition_files[1] not in CachePartitionFile.objects.all()
        )
        self.assertTrue(
            self._test_cache_partition_files[2] in CachePartitionFile.objects.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_purge_on_error(self):
        self._silence_logger(name='mayan.apps.file_caching.model_mixins')

        test_case_instance = self

        def fake_method_delete(self, *args, **kwargs):
            if self.pk == test_case_instance._test_cache_partition_files[0].pk:
                raise Exception

            return super(CachePartitionFile, self).delete(*args, **kwargs)

        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()
        self._create_test_cache_partition_file()

        self._clear_events()

        test_cache_partition_file_count = self._test_cache_partition.files.count()

        with mock.patch('mayan.apps.file_caching.models.CachePartitionFile.delete', fake_method_delete):
            self._test_cache.purge(user=self._test_case_user)

        self.assertEqual(
            self._test_cache_partition.files.count(),
            test_cache_partition_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_cache)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_cache_partition)
        self.assertEqual(events[0].verb, event_cache_partition_purged.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_cache)
        self.assertEqual(events[1].verb, event_cache_purged.id)


class CachePartitionModelTestCase(CacheTestMixin, BaseTestCase):
    def test_cache_create(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        self._clear_events()

        self._test_cache_partition.delete()

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_purge_on_error(self):
        self._silence_logger(name='mayan.apps.file_caching.model_mixins')

        test_case_instance = self

        def fake_method_delete(self, *args, **kwargs):
            if self.pk == test_case_instance._test_cache_partition_files[0].pk:
                raise Exception

            return super(CachePartitionFile, self).delete(*args, **kwargs)

        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()
        self._create_test_cache_partition_file()

        self._clear_events()

        test_cache_partition_file_count = self._test_cache_partition.files.count()

        with mock.patch('mayan.apps.file_caching.models.CachePartitionFile.delete', fake_method_delete):
            self._test_cache_partition.purge(user=self._test_case_user)

        self.assertEqual(
            self._test_cache_partition.files.count(),
            test_cache_partition_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_cache_partition)
        self.assertEqual(events[0].verb, event_cache_partition_purged.id)



from mayan.apps.appearance.classes import Icon

icon_queue_task_type_list = Icon(driver_name='fontawesome', symbol='braille')
icon_worker_list = Icon(driver_name='fontawesome', symbol='robot')
icon_worker_queue_list = Icon(driver_name='fontawesome', symbol='stream')



from mayan.apps.source_periodic.source_backend_actions.periodic_actions import SourceBackendActionPeriodicDocumentUpload


class SourceBackendActionEmailDocumentUpload(
    SourceBackendActionPeriodicDocumentUpload
):
    stored_file_identifier_name = 'message_id'



from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import (
    event_index_template_created, event_index_template_edited
)
from ..models.index_instance_models import IndexInstanceNode, IndexTemplate
from ..permissions import (
    permission_index_template_create, permission_index_template_delete,
    permission_index_template_edit, permission_index_template_rebuild
)

from .literals import (
    TEST_INDEX_TEMPLATE_LABEL, TEST_INDEX_TEMPLATE_LABEL_EDITED
)
from .mixins import (
    DocumentTypeAddRemoveIndexTemplateViewTestMixin,
    IndexTemplateNodeViewTestMixin, IndexTemplateTestMixin,
    IndexTemplateViewTestMixin
)


class DocumentTypeAddRemoveIndexTemplateViewTestCase(
    DocumentTypeAddRemoveIndexTemplateViewTestMixin, IndexTemplateTestMixin,
    GenericDocumentViewTestCase
):
    auto_add_test_index_template_to_test_document_type = False
    auto_upload_test_document = False

    def test_document_type_index_template_add_remove_get_view_no_permission(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self._clear_events()

        response = self._request_test_document_type_index_template_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_index_template),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_add_remove_get_view_with_document_type_access(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self._test_index_template),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_add_remove_get_view_with_index_template_access(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_index_template),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_add_remove_get_view_with_full_access(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_index_template),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_add_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_type_index_template_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_index_template not in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_add_view_with_document_type_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_index_template not in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_add_view_with_index_template_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_index_template not in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_add_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_index_template in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_document_type_index_template_remove_view_no_permission(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self._clear_events()

        response = self._request_test_document_type_index_template_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_index_template in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_remove_view_with_document_type_access(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_index_template in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_remove_view_with_index_template_access(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_index_template in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_index_template_remove_view_with_full_access(self):
        self._test_document_type.index_templates.add(self._test_index_template)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_document_type_index_template_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_index_template not in self._test_document_type.index_templates.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)


class IndexTemplateViewTestCase(
    IndexTemplateTestMixin, IndexTemplateViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False
    auto_create_test_index_template = False

    def test_index_template_create_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(IndexTemplate.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_create_view_with_permission(self):
        self.grant_permission(
            permission=permission_index_template_create
        )

        self._clear_events()

        response = self._request_test_index_template_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(IndexTemplate.objects.count(), 1)
        self.assertEqual(
            self._test_index_template.label, TEST_INDEX_TEMPLATE_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_created.id)

    def test_index_template_delete_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(IndexTemplate.objects.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_delete_view_with_permission(self):
        self._create_test_index_template()

        self.grant_permission(permission=permission_index_template_delete)

        self._clear_events()

        response = self._request_test_index_template_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(IndexTemplate.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_edit_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_index_template.refresh_from_db()
        self.assertEqual(
            self._test_index_template.label, TEST_INDEX_TEMPLATE_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_edit_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_index_template.refresh_from_db()
        self.assertEqual(
            self._test_index_template.label, TEST_INDEX_TEMPLATE_LABEL_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_index_template_rebuild_view_no_permission(self):
        self._create_test_document_stub()
        self._create_test_index_template()
        self._create_test_index_template_node()

        self._clear_events()

        response = self._request_test_index_template_rebuild_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(IndexInstanceNode.objects.count(), 1)
        self.assertEqual(IndexInstanceNode.objects.first().parent, None)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_rebuild_view_with_access(self):
        self._create_test_document_stub()
        self._create_test_index_template()
        self._create_test_index_template_node()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_rebuild
        )

        self._clear_events()

        response = self._request_test_index_template_rebuild_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(IndexInstanceNode.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexTemplateAddRemoveDocumentTypeViewTestCase(
    IndexTemplateTestMixin, IndexTemplateViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_index_template_document_type_add_remove_get_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_index_template),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_remove_get_view_with_document_type_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_index_template),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_remove_get_view_with_index_template_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_index_template),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_remove_get_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self._test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_index_template),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_view_no_permission(self):
        self._test_index_template.document_types.remove(
            self._test_document_type
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_view_with_document_type_access(self):
        self._test_index_template.document_types.remove(
            self._test_document_type
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_view_with_index_template_access(self):
        self._test_index_template.document_types.remove(
            self._test_document_type
        )

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_view_with_full_access(self):
        self._test_index_template.document_types.remove(
            self._test_document_type
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_index_template_document_type_remove_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_document_type_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_view_with_document_type_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_view_with_index_template_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)


class IndexTemplateNodeViewTestCase(
    IndexTemplateTestMixin, IndexTemplateNodeViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_index_template_node_create_view_no_permission(self):
        self._create_test_index_template()
        node_count = self._test_index_template.index_template_nodes.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            self._test_index_template.index_template_nodes.count(),
            node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_create_view_with_access(self):
        self._create_test_index_template()
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        node_count = self._test_index_template.index_template_nodes.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_index_template.index_template_nodes.count(),
            node_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_delete_view_no_permission(self):
        self._create_test_index_template()
        self._create_test_index_template_node()
        node_count = self._test_index_template.index_template_nodes.count()

        self._clear_events()

        response = self._request_test_index_template_node_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_index_template.index_template_nodes.count(),
            node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_delete_view_with_access(self):
        self._create_test_index_template()
        self._create_test_index_template_node()
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        node_count = self._test_index_template.index_template_nodes.count()

        self._clear_events()

        response = self._request_test_index_template_node_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_index_template.index_template_nodes.count(),
            node_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_edit_view_no_permission(self):
        self._create_test_index_template()
        self._create_test_index_template_node()
        node_expression = self._test_index_template_node.expression

        self._clear_events()

        response = self._request_test_index_template_node_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_index_template_node.refresh_from_db()
        self.assertEqual(
            self._test_index_template_node.expression, node_expression
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_edit_view_with_access(self):
        self._create_test_index_template()
        self._create_test_index_template_node()
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        node_expression = self._test_index_template_node.expression

        self._clear_events()

        response = self._request_test_index_template_node_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_index_template_node.refresh_from_db()
        self.assertNotEqual(
            self._test_index_template_node.expression, node_expression
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_list_get_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_node_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_list_get_view_with_access(self):
        self._create_test_index_template()
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_node_list_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import (
    event_signature_capture_created, event_signature_capture_deleted,
    event_signature_capture_edited
)
from ..permissions import (
    permission_signature_capture_create, permission_signature_capture_delete,
    permission_signature_capture_edit, permission_signature_capture_view
)

from .mixins import SignatureCaptureViewTestMixin


class SignatureCaptureViewTestCase(
    SignatureCaptureViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_signature_capture_create_view_no_permission(self):
        signature_capture_count = self._test_document.signature_captures.count()

        self._clear_events()

        response = self._request_test_signature_capture_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_create_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_create
        )

        signature_capture_count = self._test_document.signature_captures.count()

        self._clear_events()

        response = self._request_test_signature_capture_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_signature_capture)
        self.assertEqual(events[0].verb, event_signature_capture_created.id)

    def test_trashed_document_signature_capture_create_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_create
        )

        signature_capture_count = self._test_document.signature_captures.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_delete_view_no_permission(self):
        self._create_test_signature_capture()

        signature_capture_count = self._test_document.signature_captures.count()

        self._clear_events()

        response = self._request_test_signature_capture_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_delete_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_delete
        )

        signature_capture_count = self._test_document.signature_captures.count()

        self._clear_events()

        response = self._request_test_signature_capture_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_signature_capture_deleted.id)

    def test_signature_capture_delete_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_delete
        )

        signature_capture_count = self._test_document.signature_captures.count()

        self._clear_events()

        response = self._request_test_signature_capture_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_signature_capture_deleted.id)

    def test_trashed_document_signature_capture_delete_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_delete
        )

        signature_capture_count = self._test_document.signature_captures.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_capture_delete_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_delete
        )

        signature_capture_count = self._test_document.signature_captures.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.signature_captures.count(),
            signature_capture_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_detail_view_no_permission(self):
        self._create_test_signature_capture()

        self._clear_events()

        response = self._request_test_signature_capture_detail_view()
        self.assertEqual(response.status_code, 404)

        self._test_signature_capture.refresh_from_db()
        self.assertNotContains(
            response=response, text=self._test_signature_capture.text,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_detail_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_view
        )

        self._clear_events()

        response = self._request_test_signature_capture_detail_view()
        self.assertEqual(response.status_code, 200)

        self._test_signature_capture.refresh_from_db()
        self.assertContains(
            response=response, text=self._test_signature_capture.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_detail_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_view
        )

        self._clear_events()

        response = self._request_test_signature_capture_detail_view()
        self.assertEqual(response.status_code, 200)

        self._test_signature_capture.refresh_from_db()
        self.assertContains(
            response=response, text=self._test_signature_capture.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_capture_detail_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_detail_view()
        self.assertEqual(response.status_code, 404)

        self._test_signature_capture.refresh_from_db()
        self.assertNotContains(
            response=response, text=self._test_signature_capture.text,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_capture_detail_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_detail_view()
        self.assertEqual(response.status_code, 404)

        self._test_signature_capture.refresh_from_db()
        self.assertNotContains(
            response=response, text=self._test_signature_capture.text,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_edit_view_no_permission(self):
        self._create_test_signature_capture()

        signature_capture_text = self._test_signature_capture.text

        self._clear_events()

        response = self._request_test_signature_capture_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_signature_capture.refresh_from_db()
        self.assertEqual(
            self._test_signature_capture.text, signature_capture_text
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_edit_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_edit
        )

        signature_capture_text = self._test_signature_capture.text

        self._clear_events()

        response = self._request_test_signature_capture_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_signature_capture.refresh_from_db()
        self.assertNotEqual(
            self._test_signature_capture.text, signature_capture_text
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_signature_capture)
        self.assertEqual(events[0].verb, event_signature_capture_edited.id)

    def test_signature_capture_edit_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_edit
        )

        signature_capture_text = self._test_signature_capture.text

        self._clear_events()

        response = self._request_test_signature_capture_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_signature_capture.refresh_from_db()
        self.assertNotEqual(
            self._test_signature_capture.text, signature_capture_text
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_signature_capture)
        self.assertEqual(events[0].verb, event_signature_capture_edited.id)

    def test_trashed_document_signature_capture_edit_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_document,
            permission=permission_signature_capture_edit
        )

        signature_capture_text = self._test_signature_capture.text

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_signature_capture.refresh_from_db()
        self.assertEqual(
            self._test_signature_capture.text, signature_capture_text
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_capture_edit_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_edit
        )

        signature_capture_text = self._test_signature_capture.text

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_signature_capture.refresh_from_db()
        self.assertEqual(
            self._test_signature_capture.text, signature_capture_text
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_list_view_with_no_permission(self):
        self._create_test_signature_capture()

        self._clear_events()

        response = self._request_test_signature_capture_list_view()
        self.assertNotContains(
            response=response, text=self._test_signature_capture.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_list_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_view
        )

        self._clear_events()

        response = self._request_test_signature_capture_list_view()
        self.assertContains(
            response=response, text=self._test_signature_capture.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_capture_list_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_view
        )

        self._clear_events()

        response = self._request_test_signature_capture_list_view()
        self.assertContains(
            response=response, text=self._test_signature_capture.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_capture_list_view_with_document_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_list_view()
        self.assertNotContains(
            response=response, text=self._test_signature_capture.text,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_capture_list_view_with_signature_capture_access(self):
        self._create_test_signature_capture()

        self.grant_access(
            obj=self._test_signature_capture,
            permission=permission_signature_capture_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_signature_capture_list_view()
        self.assertNotContains(
            response=response, text=self._test_signature_capture.text,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from django.contrib.auth import get_user_model
from django.utils.deprecation import MiddlewareMixin

from mayan.apps.acls.models import AccessControlList
from mayan.apps.user_management.querysets import get_user_queryset

from ..events import (
    event_user_impersonation_ended, event_user_impersonation_started
)
from ..literals import (
    USER_IMPERSONATE_VARIABLE_DISABLE, USER_IMPERSONATE_VARIABLE_ID,
    USER_IMPERSONATE_VARIABLE_PERMANENT
)
from ..permissions import permission_users_impersonate

logger = logging.getLogger(name=__name__)


class ImpersonateMiddleware(MiddlewareMixin):
    @staticmethod
    def get_user(pk, request):
        queryset = AccessControlList.objects.restrict_queryset(
            queryset=get_user_queryset().exclude(pk=request.user.pk),
            permission=permission_users_impersonate, user=request.user
        )

        return queryset.get(pk=pk)

    def process_request(self, request):
        User = get_user_model()

        impersonate_permanent_session = USER_IMPERSONATE_VARIABLE_PERMANENT in request.session

        if not impersonate_permanent_session:
            if USER_IMPERSONATE_VARIABLE_DISABLE in request.POST or USER_IMPERSONATE_VARIABLE_DISABLE in request.GET:
                # End the impersonation
                if USER_IMPERSONATE_VARIABLE_ID in request.session:
                    user_impersonate_id = request.session[USER_IMPERSONATE_VARIABLE_ID]

                    del request.session[USER_IMPERSONATE_VARIABLE_ID]

                    try:
                        user = ImpersonateMiddleware.get_user(
                            pk=user_impersonate_id, request=request
                        )
                    except User.DoesNotExist:
                        return
                    else:
                        event_user_impersonation_ended.commit(
                            actor=request.user, target=user
                        )
                        return
            else:
                # Start the impersonation
                user_impersonate_id = request.GET.get(
                    USER_IMPERSONATE_VARIABLE_ID, request.POST.get(
                        USER_IMPERSONATE_VARIABLE_ID
                    )
                )
                user_impersonate_permanent = USER_IMPERSONATE_VARIABLE_PERMANENT in request.GET or USER_IMPERSONATE_VARIABLE_PERMANENT in request.POST

                if user_impersonate_id:
                    try:
                        user_impersonate_id = int(user_impersonate_id)
                    except ValueError:
                        logger.error(
                            'Unable to impersonate, invalid user ID value. %s',
                            user_impersonate_id
                        )
                    else:
                        try:
                            user = ImpersonateMiddleware.get_user(
                                pk=user_impersonate_id, request=request
                            )
                        except User.DoesNotExist:
                            return
                        else:
                            request.session[USER_IMPERSONATE_VARIABLE_ID] = user_impersonate_id

                            event_user_impersonation_started.commit(
                                actor=request.user, target=user
                            )

                            request.user = user

                            if user_impersonate_permanent:
                                request.session[USER_IMPERSONATE_VARIABLE_PERMANENT] = True

                            return

        # Set the request user to the previously impersonated user.
        if USER_IMPERSONATE_VARIABLE_ID in request.session:
            user_impersonate_id = request.session[USER_IMPERSONATE_VARIABLE_ID]
            try:
                user = ImpersonateMiddleware.get_user(
                    pk=user_impersonate_id, request=request
                )
            except User.DoesNotExist:
                del request.session[USER_IMPERSONATE_VARIABLE_ID]
                return
            else:
                request.user = user



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('logging', '0001_initial')
    ]

    operations = [
        migrations.RenameModel(
            old_name='ErrorLog',
            new_name='StoredErrorLog'
        )
    ]



from django.apps import apps
from django.contrib.staticfiles.management.commands.collectstatic import (
    Command as DjangoCommand
)


class Command(DjangoCommand):
    help = 'Call the collectstatic command with some specific defaults.'

    def handle(self, **options):
        """
        Collect static_media_ignore_patterns from all apps. The pattern
        matches anything after "/<app name>/static/"
        """
        self.verbosity = options['verbosity']

        for key, data in apps.app_configs.items():
            options['ignore_patterns'].extend(
                getattr(
                    data, 'static_media_ignore_patterns', ()
                )
            )

        if options['verbosity'] >= 1:
            self.log(
                'Ignore patterns: {}'.format(
                    options['ignore_patterns']
                ), level=1
            )

        return super().handle(**options)



from rest_framework import status

from mayan.apps.django_gpg.permissions import permission_key_sign
from mayan.apps.django_gpg.tests.literals import TEST_KEY_PUBLIC_ID
from mayan.apps.django_gpg.tests.mixins import KeyTestMixin
from mayan.apps.documents.events import (
    event_document_file_created, event_document_file_edited,
    event_document_version_created, event_document_version_edited,
    event_document_version_page_created
)
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_embedded_signature_created
from ..permissions import (
    permission_document_file_sign_embedded,
    permission_document_file_signature_view
)

from .literals import TEST_SIGNED_DOCUMENT_PATH
from .mixins import EmbeddedSignatureAPIViewTestMixin


class EmbeddedSignatureDocumentAPIViewTestCase(
    DocumentTestMixin, EmbeddedSignatureAPIViewTestMixin,
    KeyTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def test_document_signature_embedded_detail_api_view_no_permission(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self._clear_events()

        response = self._request_test_document_signature_embedded_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_embedded_detail_api_view_with_access(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._clear_events()

        response = self._request_test_document_signature_embedded_detail_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['key_id'], TEST_KEY_PUBLIC_ID
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_embedded_detail_api_view_with_access(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_embedded_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_embedded_list_api_view_no_permission(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self._clear_events()

        response = self._request_test_document_signature_embedded_list_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_embedded_list_api_view_with_access(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._clear_events()

        response = self._request_test_document_signature_embedded_list_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['key_id'], TEST_KEY_PUBLIC_ID
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_embedded_list_api_view_with_access(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_embedded_list_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_embedded_sign_api_view_with_no_permission(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self._clear_events()

        response = self._request_test_document_signature_embedded_sign_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_embedded_sign_api_view_with_document_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_embedded
        )

        self._clear_events()

        response = self._request_test_document_signature_embedded_sign_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_embedded_sign_api_view_with_key_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_signature_embedded_sign_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_embedded_sign_api_view_with_full_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_embedded
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_signature_embedded_sign_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document_file = self._test_document.file_latest
        test_document_version = self._test_document.versions.last()

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, test_document_version.pages.first()
        )
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[4].action_object, self._test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version)
        self.assertEqual(events[4].verb, event_document_version_edited.id)

        self.assertEqual(
            events[5].action_object,
            self._test_document.file_latest.signatures.first().embeddedsignature
        )
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_file)
        self.assertEqual(events[5].verb, event_embedded_signature_created.id)

    def test_trashed_document_signature_embedded_sign_api_view_with_full_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_embedded
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_embedded_sign_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_document_version_page_created, event_document_version_page_deleted,
    event_document_version_page_edited
)
from ..permissions import (
    permission_document_version_edit, permission_document_version_view
)

from .mixins.document_mixins import DocumentTestMixin
from .mixins.document_version_mixins import DocumentVersionPageAPIViewTestMixin


class DocumentVersionPageAPIViewTestCase(
    DocumentVersionPageAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    def test_document_version_page_create_api_view_no_permission(self):
        document_version_count_page = self._test_document.version_active.version_pages.count()

        self._clear_events()

        response = self._request_test_document_version_page_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.version_active.version_pages.count(),
            document_version_count_page
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_create_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_count_page = self._test_document.version_active.version_pages.count()

        self._clear_events()

        response = self._request_test_document_version_page_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            self._test_document.version_active.version_pages.count(),
            document_version_count_page + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_version)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version_page)
        self.assertEqual(events[0].verb, event_document_version_page_created.id)

    def test_trashed_document_version_page_create_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_count_page = self._test_document.version_active.version_pages.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.version_active.version_pages.count(),
            document_version_count_page
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_delete_api_view_no_permission(self):
        document_version_count_page = self._test_document.version_active.version_pages.count()

        self._clear_events()

        response = self._request_test_document_version_page_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.version_active.version_pages.count(),
            document_version_count_page
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_delete_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_count_page = self._test_document.version_active.version_pages.count()

        self._clear_events()

        response = self._request_test_document_version_page_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            self._test_document.version_active.version_pages.count(),
            document_version_count_page - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version)
        self.assertEqual(events[0].verb, event_document_version_page_deleted.id)

    def test_trashed_document_version_page_delete_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_count_page = self._test_document.version_active.version_pages.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.version_active.version_pages.count(),
            document_version_count_page
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_edit_via_patch_api_view_no_permission(self):
        document_version_page_number = self._test_document_version_page.page_number

        self._clear_events()

        response = self._request_test_document_version_page_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_version_page.refresh_from_db()
        self.assertEqual(
            self._test_document_version_page.page_number,
            document_version_page_number
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_edit_via_patch_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_page_number = self._test_document_version_page.page_number

        self._clear_events()

        response = self._request_test_document_version_page_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_version_page.refresh_from_db()
        self.assertNotEqual(
            self._test_document_version_page.page_number,
            document_version_page_number
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_version)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version_page)
        self.assertEqual(events[0].verb, event_document_version_page_edited.id)

    def test_trashed_document_version_page_edit_via_patch_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_page_number = self._test_document_version_page.page_number

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_version_page.refresh_from_db()
        self.assertEqual(
            self._test_document_version_page.page_number,
            document_version_page_number
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_edit_via_put_api_view_no_permission(self):
        document_version_page_number = self._test_document_version_page.page_number

        self._clear_events()

        response = self._request_test_document_version_page_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_version_page.refresh_from_db()
        self.assertEqual(
            self._test_document_version_page.page_number,
            document_version_page_number
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_edit_via_put_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_page_number = self._test_document_version_page.page_number

        self._clear_events()

        response = self._request_test_document_version_page_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_version_page.refresh_from_db()
        self.assertNotEqual(
            self._test_document_version_page.page_number,
            document_version_page_number
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(events[0].action_object, self._test_document_version)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version_page)
        self.assertEqual(events[0].verb, event_document_version_page_edited.id)

    def test_trashed_document_version_page_edit_via_put_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        document_version_page_number = self._test_document_version_page.page_number

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_version_page.refresh_from_db()
        self.assertEqual(
            self._test_document_version_page.page_number,
            document_version_page_number
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_image_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_image_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_image_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_document_version_page.id
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.contrib.contenttypes.models import ContentType

from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from .mixins import CommonAPITestMixin


class CommonAPITestCase(CommonAPITestMixin, BaseAPITestCase):
    auto_login_user = False

    def test_content_type_detail_api_view(self):
        self._clear_events()

        response = self._request_content_type_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.json()['id'], self._test_content_type.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_content_type_list_api_view(self):
        self._clear_events()

        response = self._request_content_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.json()['count'], ContentType.objects.count()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import itertools

from django.core.exceptions import ValidationError
from django.db import connection, models
from django.db.models import Q, Value
from django.db.models.functions import Cast, Replace

from ..exceptions import DynamicSearchValueTransformationError
from ..search_backends import SearchBackend
from ..search_fields import SearchFieldVirtualAllFields
from ..search_models import SearchModel
from ..search_query_types import (
    BackendQueryType, QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
    QueryTypeLessThanOrEqual, QueryTypePartial, QueryTypeRange,
    QueryTypeRangeExclusive, QueryTypeRegularExpression
)

from .literals.django_literals import (
    DEFAULT_FUZZY_SLOP, DJANGO_TO_DJANGO_FIELD_MAP, MAXIMUM_FUZZY_OPTIONS
)


class BackendQueryTypeExact(BackendQueryType):
    query_type = QueryTypeExact

    def do_resolve(self):
        if self.value is not None:
            if self.get_search_backend_field_type() == models.BooleanField:
                lookup_template = '{field_name}__exact'
                value_template = '{}'
            elif self.get_search_backend_field_type() == models.DateTimeField:
                backend_query_type = BackendQueryTypeRange(
                    is_quoted_value=self.is_quoted_value,
                    search_backend=self.search_backend,
                    search_field=self.search_field, value=(
                        self.value, self.value.replace(microsecond=999999)
                    ), extra_kwargs=self.extra_kwargs
                )
                return backend_query_type.do_resolve()
            elif self.get_search_backend_field_type() == models.UUIDField:
                lookup_template = '{field_name}_clean__exact'
                value_template = '{}'
            elif self.get_search_backend_field_type() == models.IntegerField:
                lookup_template = '{field_name}__exact'
                value_template = '{}'
            elif self.get_search_backend_field_type() == models.PositiveIntegerField:
                lookup_template = '{field_name}__exact'
                value_template = '{}'
            else:
                if self.is_quoted_value and self.value == '':
                    lookup_template = '{field_name}__exact'
                    value_template = '{}'
                else:
                    lookup_template = '{field_name}_clean__iregex'
                    if connection.vendor == 'postgresql':
                        value_template = r'\y{}\y'
                    else:
                        value_template = r'\b{}\b'

            return Q(
                **{
                    lookup_template.format(
                        field_name=self.search_field.field_name
                    ): value_template.format(self.value)
                }
            )


class BackendQueryFuzzy(BackendQueryType):
    query_type = QueryTypeFuzzy

    def do_resolve(self):
        fuzzy_options = []

        if self.value is not None:
            permutation_list = list(
                set(
                    [
                        ''.join(letters) for letters in itertools.permutations(self.value)
                    ]
                )
            )

            for permutation in permutation_list:
                difference_count = sum(1 for a, b in zip(self.value, permutation) if a != b)

                if difference_count <= DEFAULT_FUZZY_SLOP:
                    fuzzy_options.append(permutation)

            result = None
            for entry in fuzzy_options[:MAXIMUM_FUZZY_OPTIONS]:
                backend_query_type = BackendQueryTypeExact(
                    is_quoted_value=self.is_quoted_value,
                    search_backend=self.search_field,
                    search_field=self.search_field, value=entry,
                    extra_kwargs=self.extra_kwargs
                )

                query = backend_query_type.do_resolve()

                if result is None:
                    result = query
                else:
                    result |= query

            return result


class BackendQueryTypeGreaterThan(BackendQueryType):
    query_type = QueryTypeGreaterThan

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}__gt'.format(
                        field_name=self.search_field.field_name
                    ): self.value
                }
            )


class BackendQueryTypeGreaterThanOrEqual(BackendQueryType):
    query_type = QueryTypeGreaterThanOrEqual

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}__gte'.format(
                        field_name=self.search_field.field_name
                    ): self.value
                }
            )


class BackendQueryTypeLessThan(BackendQueryType):
    query_type = QueryTypeLessThan

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}__lt'.format(
                        field_name=self.search_field.field_name
                    ): self.value
                }
            )


class BackendQueryTypeLessThanOrEqual(BackendQueryType):
    query_type = QueryTypeLessThanOrEqual

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}__lte'.format(
                        field_name=self.search_field.field_name
                    ): self.value
                }
            )


class BackendQueryTypePartial(BackendQueryType):
    query_type = QueryTypePartial

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}_clean__icontains'.format(
                        field_name=self.search_field.field_name
                    ): self.value
                }
            )


class BackendQueryTypeRange(BackendQueryType):
    query_type = QueryTypeRange

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}__gte'.format(
                        field_name=self.search_field.field_name
                    ): self.value[0], '{field_name}__lte'.format(
                        field_name=self.search_field.field_name
                    ): self.value[1]
                }
            )


class BackendQueryTypeRangeExclusive(BackendQueryType):
    query_type = QueryTypeRangeExclusive

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}__gt'.format(
                        field_name=self.search_field.field_name
                    ): self.value[0], '{field_name}__lt'.format(
                        field_name=self.search_field.field_name
                    ): self.value[1]
                }
            )


class BackendQueryTypeRegularExpression(BackendQueryType):
    query_type = QueryTypeRegularExpression

    def do_resolve(self):
        if self.value is not None:
            return Q(
                **{
                    '{field_name}_clean__regex'.format(
                        field_name=self.search_field.field_name
                    ): self.value
                }
            )


class DjangoSearchBackend(SearchBackend):
    field_type_mapping = DJANGO_TO_DJANGO_FIELD_MAP

    def _do_search_model_filter(self, filter_kwargs, limit, search_field):
        if search_field.field_class == models.UUIDField:
            # Remove hyphens when searching UUID fields.
            replace_function = Replace(
                expression=Cast(
                    expression=search_field.field_name,
                    output_field=models.CharField()
                ), text=Value('-'), replacement=Value(''),
                output_field=models.CharField()
            )
        else:
            replace_function = Replace(
                expression=Cast(
                    expression=search_field.field_name,
                    output_field=models.CharField()
                ), text=Value('-'), replacement=Value('_'),
                output_field=models.CharField()
            )

        queryset = search_field.search_model.get_queryset().annotate(
            **{
                '{}_clean'.format(search_field.field_name): replace_function
            }
        )

        try:
            return tuple(
                set(
                    queryset.filter(
                        filter_kwargs
                    ).values_list('pk', flat=True)
                )
            )[0:limit]
        except ValidationError:
            return ()

    def _get_status(self):
        result = []

        for search_model in SearchModel.all():
            queryset = search_model.get_queryset()

            result.append(
                '{}: {}'.format(
                    search_model.label, queryset.count()
                )
            )

        return '\n'.join(result)

    def _search(
        self, limit, search_field, query_type, value, is_quoted_value=False,
        is_raw_value=False
    ):
        self.do_query_type_verify(
            query_type=query_type, search_field=search_field
        )

        if isinstance(search_field, SearchFieldVirtualAllFields):
            result = set()

            for search_field in search_field.field_composition:
                try:
                    search_field_query = query_type.resolve_for_backend(
                        is_quoted_value=is_quoted_value,
                        is_raw_value=is_raw_value, search_backend=self,
                        search_field=search_field, value=value
                    )
                except DynamicSearchValueTransformationError:
                    """Skip the search field."""
                else:
                    if search_field_query is not None:
                        field_id_list = self._do_search_model_filter(
                            filter_kwargs=search_field_query, limit=limit,
                            search_field=search_field
                        )

                        result.update(field_id_list)

            return result
        else:
            try:
                filter_kwargs = query_type.resolve_for_backend(
                    is_quoted_value=is_quoted_value,
                    is_raw_value=is_raw_value, search_backend=self,
                    search_field=search_field, value=value
                )
            except DynamicSearchValueTransformationError:
                return ()
            else:
                if filter_kwargs is None:
                    return ()
                else:
                    return self._do_search_model_filter(
                        filter_kwargs=filter_kwargs, limit=limit,
                        search_field=search_field
                    )


BackendQueryType.register(
    klass=BackendQueryTypeExact, search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryFuzzy, search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThan, search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThanOrEqual,
    search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThan, search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThanOrEqual,
    search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypePartial, search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRange,
    search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRangeExclusive,
    search_backend=DjangoSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRegularExpression,
    search_backend=DjangoSearchBackend
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Converter'), name='converter'
)

event_asset_created = namespace.add_event_type(
    label=_(message='Asset created'), name='asset_created'
)
event_asset_edited = namespace.add_event_type(
    label=_(message='Asset edited'), name='asset_edited'
)



from .document_file_models import *  # NOQA
from .document_file_page_models import *  # NOQA
from .document_models import *  # NOQA
from .document_type_models import *  # NOQA
from .document_version_models import *  # NOQA
from .document_version_page_models import *  # NOQA
from .favorite_document_models import *  # NOQA
from .recently_accessed_document_models import *  # NOQA
from .trashed_document_models import *  # NOQA



from django.contrib import admin

from .models import Theme, UserThemeSetting


@admin.register(Theme)
class ThemeAdmin(admin.ModelAdmin):
    list_display = ('label',)
    search_fields = ('label',)


@admin.register(UserThemeSetting)
class UserThemeSettingAdmin(admin.ModelAdmin):
    list_display = ('user', 'theme',)
    list_filter = ('theme__label',)
    search_fields = (
        'user__username', 'user__first_name', 'user__last_name',
        'user__email', 'theme__label'
    )



import logging
from pathlib import Path

from django.apps import apps
from django.core.files import File
from django.utils.translation import gettext, gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.converter.exceptions import AppImageError
from mayan.apps.storage.compressed_files import Archive
from mayan.apps.storage.exceptions import NoMIMETypeMatch

from ..events import event_document_type_changed
from ..literals import (
    DEFAULT_DOCUMENT_FILE_ACTION_NAME, IMAGE_ERROR_NO_ACTIVE_VERSION
)
from ..permissions import permission_document_change_type
from ..signals import signal_post_document_type_change

from .document_type_models import DocumentType

logger = logging.getLogger(name=__name__)


class DocumentBusinessLogicMixin:
    @classmethod
    def execute_pre_create_hooks(cls, kwargs=None):
        """
        Helper method to allow checking if it is possible to create
        a new document.
        """
        cls._execute_hooks(
            hook_list=cls._hooks_pre_create, kwargs=kwargs
        )

    @classmethod
    def register_pre_create_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._hooks_pre_create, func=func, order=order
        )

    def _document_type_change(self, document_type, force=False, user=None):
        has_changed = self.document_type != document_type

        if has_changed or force:
            self.document_type = document_type

            self._event_ignore = True
            self.save(
                update_fields=('document_type',)
            )

            if user:
                self.add_as_recent_document_for_user(user=user)

            signal_post_document_type_change.send(
                sender=self.__class__, instance=self
            )

            event_document_type_changed.commit(
                action_object=document_type, actor=user, target=self
            )

    def add_as_recent_document_for_user(self, user):
        RecentlyAccessedDocument = apps.get_model(
            app_label='documents', model_name='RecentlyAccessedDocument'
        )
        return RecentlyAccessedDocument.valid.add_document_for_user(
            document=self, user=user
        )

    def document_type_change(self, document_type, user, force=False):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        queryset = AccessControlList.objects.restrict_queryset(
            queryset=Document.valid.all(),
            permission=permission_document_change_type,
            user=user
        )

        # Verify the user has the access to change the document's type.
        queryset.get(pk=self.pk)

        queryset = AccessControlList.objects.restrict_queryset(
            queryset=DocumentType.objects.all(),
            permission=permission_document_change_type,
            user=user
        )

        # Verify the user has the access to change into the new document
        # type.
        document_type = queryset.get(pk=document_type.pk)

        return self._document_type_change(
            document_type=document_type, force=force, user=user
        )

    def files_upload(
        self, file_object, action_name=None, comment=None, filename=None,
        expand=False, user=None
    ):
        logger.info('Creating new document file for document: %s', self)

        if not action_name:
            action_name = DEFAULT_DOCUMENT_FILE_ACTION_NAME

        if not comment:
            comment = ''

        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        if expand:
            try:
                compressed_file = Archive.open(file_object=file_object)
                for compressed_file_member in compressed_file.members():
                    with compressed_file.open_member(filename=compressed_file_member) as compressed_file_member_file_object:
                        # Recursive call to expand nested compressed files
                        # expand=True literal for recursive nested files.
                        # Might cause problem with office files inside a
                        # compressed file.
                        # Don't use keyword arguments for Path to allow
                        # partials.
                        self.file_upload(
                            action_name=action_name, comment=comment,
                            expand=False,
                            file_object=compressed_file_member_file_object,
                            filename=Path(compressed_file_member).name,
                            user=user
                        )

                # Avoid executing the expand=False code path.
                return
            except NoMIMETypeMatch:
                logger.debug(msg='No expanding; Exception: NoMIMETypeMatch')
                # Fall through to same code path as expand=False to avoid
                # duplicating code.

        try:
            filename = filename or Path(file_object.name).name

            document_file = DocumentFile(
                document=self, comment=comment, file=File(file=file_object),
                filename=filename
            )
            document_file._event_actor = user
            document_file.save()
        except Exception as exception:
            logger.error(
                'Error creating new file for document: %s; %s', self,
                exception, exc_info=True
            )
            raise
        else:
            logger.info('New document file queued for document: %s', self)

            document_file.versions_new(
                action_name=action_name, comment=comment, user=user
            )

    def get_api_image_url(self, *args, **kwargs):
        version_active = self.version_active
        if version_active:
            return version_active.get_api_image_url(*args, **kwargs)
        else:
            raise AppImageError(error_name=IMAGE_ERROR_NO_ACTIVE_VERSION)

    def get_label(self):
        return self.label or gettext(
            message='Document stub, id: %d'
        ) % self.pk

    get_label.short_description = _(message='Label')

    @property
    def is_in_trash(self):
        return self.in_trash

    @property
    def pages(self):
        try:
            return self.version_active.pages
        except AttributeError:
            # Document has no version yet.
            DocumentVersionPage = apps.get_model(
                app_label='documents', model_name='DocumentVersionPage'
            )

            return DocumentVersionPage.objects.none()



from django.core.files.base import ContentFile

from mayan.apps.sources.source_backends.base import SourceBackend

from ..source_backend_actions.periodic_actions import SourceBackendActionPeriodicDocumentUpload
from ..source_backends.mixins import SourceBackendMixinPeriodic


class SourceBackendTestPeriodic(SourceBackendMixinPeriodic, SourceBackend):
    action_class_list = (SourceBackendActionPeriodicDocumentUpload,)
    label = 'Test periodic source backend'

    def action_file_delete(self, encoded_filename):
        return

    def action_file_get(self, encoded_filename):
        yield {
            'file': ContentFile('')
        }

    def get_file_identifier(self):
        return ''

    def get_stored_file_list(self):
        yield ''



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('cabinets', '0005_auto_20210525_0500')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cabinet',
            options={
                'verbose_name': 'Cabinet', 'verbose_name_plural': 'Cabinets'
            }
        )
    ]



from rest_framework import status

from mayan.apps.messaging.events import event_message_created
from mayan.apps.messaging.models import Message
from mayan.apps.rest_api.tests.base import BaseAPITransactionTestCase
from mayan.apps.storage.events import event_download_file_created
from mayan.apps.storage.models import DownloadFile

from ..events import event_document_version_exported
from ..permissions import permission_document_version_export

from .mixins import DocumentVersionExportAPIViewTestMixin


class DocumentVersionExportAPIViewTestCase(
    DocumentVersionExportAPIViewTestMixin, BaseAPITransactionTestCase
):
    def test_document_version_export_api_view_via_get_no_permission(self):
        download_file_count = DownloadFile.objects.count()
        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_document_version_export_api_view_via_get()
        self.assertEqual(
            response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
        )

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )
        self.assertEqual(Message.objects.count(), message_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_export_api_view_via_get_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_export
        )
        download_file_count = DownloadFile.objects.count()
        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_document_version_export_api_view_via_get()
        self.assertEqual(
            response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
        )

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )
        self.assertEqual(Message.objects.count(), message_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_export_api_view_via_get_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_export
        )
        download_file_count = DownloadFile.objects.count()
        message_count = Message.objects.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_export_api_view_via_get()
        self.assertEqual(
            response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
        )

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )
        self.assertEqual(Message.objects.count(), message_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_export_api_view_via_post_no_permission(self):
        download_file_count = DownloadFile.objects.count()
        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_document_version_export_api_view_via_post()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )
        self.assertEqual(Message.objects.count(), message_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_export_api_view_via_post_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_export
        )
        download_file_count = DownloadFile.objects.count()
        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_document_version_export_api_view_via_post()
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count + 1
        )
        self.assertEqual(Message.objects.count(), message_count + 1)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_download_file)
        self.assertEqual(events[0].verb, event_download_file_created.id)

        self.assertEqual(events[1].action_object, test_download_file)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_version)
        self.assertEqual(events[1].verb, event_document_version_exported.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, test_message)
        self.assertEqual(events[2].target, test_message)
        self.assertEqual(events[2].verb, event_message_created.id)

    def test_trashed_document_version_export_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_export
        )
        download_file_count = DownloadFile.objects.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_export_api_view_via_post()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)
